// fixbbl.c : Fix the bbl file to support GBK characters (displaying, sorting, etc).
// author:  hooklee (Shujun Li) in 2003
// reach hooklee at http://www.hooklee.com or www.chinatex.org

#define VERSION "0.41"
#define TODAY "30 November 2005"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#ifdef WIN32
#  include <io.h>
#  define PATH_MAX	_MAX_PATH+1
#else
#  include <unistd.h>
#  include <limits.h>
#  include <glob.h>
#endif

#define BYTE unsigned char
#define DWORD unsigned int

//is a valid high byte of some GBK character
#define GBK_HIGH(h)  ( 0x81<=(h&0xff) && (h&0xff)<=0xfe )

//is a valid low byte of some GBK character
#define GBK_LOW(l)  ( 0x40<=(l&0xff) && (l&0xff)!=0x7f && (l&0xff)!=0xff)

#ifndef WIN32
#  define _fileno(f) f
static size_t _filelength(FILE *f)
{
    size_t pos = ftell(f);
    size_t length;

    fseek(f, 0, SEEK_END);
    length = ftell(f);
    fseek(f, pos, SEEK_SET);
    return length;
}
#endif

#if !defined(WIN32) && !defined(GO32)
static int strnicmp (const char *s0, const char *s1, int n)
{
    int i;
    while (n-- > 0 && *s0 != '\0' && *s1 != '\0') {
	i = toupper(*(s0++)) - toupper(*(s1++));
	if (i) return i;
    }
    return n > 0 ? toupper(*s0) - toupper(*s1) : 0;
}
#endif

//is a character c '0'...'9','a'...'f','A'...'F'?
int is16digit(char c)
{
	return isdigit(c) || (c>='a' && c<='f') || (c>='A' && c<='F') ? 1 : 0;
}

//2-digit hexadecimal string to decimal number
unsigned char xtoi(char *str)
{
	unsigned char h,l;
	h = (unsigned char)tolower(*str);
	l = (unsigned char)tolower(*(str+1));
	if isdigit(h) h = h - '0';
	else h = h - 'a' + 10;
	if isdigit(l) l = l - '0';
	else l = l - 'a' + 10;
	return 16*h + l;
}

void version(void)
{
  printf("fixbbl, version "VERSION", by hooklee on "TODAY"\n");
}

void discription()
{
  printf("This tool fixes .bbl file to solve problems with CJK and BibTeX as follows:\n");
  printf("1) transform CJK characters to '^^7f'-format;\n");
  printf("2) transform English keywords to Chinese:\n");
  printf("   \"\\&\", \"and\", \"et al.\", \"et~al.\", \"In \", \", editors\", \"pages \", \"chapter \" and ");
  printf("     \" edition\" (default Chinese keywords can be found in fixbbl.cfg);\n");
  printf("3) sort Chinese bibitems by PINYIN (ƴ), BIHUA (ʻ) or Unicode.\n");
  printf("For more information, please see fixbbl.pdf.\n");
  printf("Please visit http://www.hooklee.com to report bugs and comments.\n");
}

void CJKcommands()
{
  printf("Place the following commands in preamble of your .tex file\n  or directly modify your CJK.sty in your local TeX system.\n");
  printf("With the help of the commands, problems with .bbl can be partially solved.\n");
  printf("Note: The commands will be officially released in future versions of CJK.\n");
  printf("      They cannot transform English terms to Chinese terms.\n");
  printf("\n");
  printf("\\makeatletter\n");
  printf("\\AtBeginDocument{%%\n");
  printf("  \\if@filesw\n");
  printf("    \\immediate\\write\\@mainaux{%%\n");
  printf("      \\string\\begin{CJK}{GBK}{song}}%%\n");
  printf("    \\immediate\\write\\@mainaux{%%\n");
  printf("      \\string\\makeatletter}%%\n");
  printf("  \\fi}\n");
  printf("\\AtEndDocument{%%\n");
  printf("  \\if@filesw\n");
  printf("    \\immediate\\write\\@mainaux{%%\n");
  printf("      \\string\\end{CJK}}%%\n");
  printf("  \\fi}\n");
  printf("\\makeatother\n");
}

void RedefinableStrings()
{
  version();
  printf("The following are commands to redefine strings:\n");
  printf("They can also be pre-stored in fixbbl.cfg ('-' is cancelled).\n");
  printf("********************************************************************************");
  printf("\"-and=, \"\t\t redefine the replacing string of \"\\&\" and \"and\"\n");
  printf("\"-etal.=\"\t\t redefine the replacing string of \"et~al.\" and \"et al.\"\n");
  printf("\"-_In=In \"\t\t redefine the replaced string \"In \"\n");
  printf("\"-In=: \"\t\t redefine the replacing string of \"In \"\n");
  printf("\"-_pages=pages \"\t redefine the replaced string \"pages \"\n");
  printf("\"-pages=ҳ: \"\t\t redefine the replacing string \"pages \"\n");
  printf("\"-_chapter=chapter \"\t redefine the replaced string \"chapter \"\n");
  printf("\"-chapter=\"\t\t redefine the replacing string \"chapter \"\n");
  printf("\"-_edition= edition\"\t redefine the replaced string \" edition\"\n");
  printf("\"-edition=\"\t\t redefine the replacing string \" edition\"\n");
  printf("\"-_editors=, editors\"\t redefine the replaced string \", editors\" in @Proceedings item\n");
  printf("\"-_editor=, editor\"\t redefine the replaced string \", editor\" in @Proceedings item\n");
  printf("\"-editors= ()\"\t redefine the replacing string of \", editor(s)\" in @Proceedings item\n");
  printf("\"-_editors2=, editors\"\t redefine the replaced string \", editors\" in @InProceedings item\n");
  printf("\"-_editor2=, editor\"\t redefine the replaced string \", editor\" in @InProceedings item\n");
  printf("\"-editors2=\"\t redefine the replacing string of \", editor(s)\" in @InProceedings item\n");
  printf("********************************************************************************");
}

//print usage of gbk2uni
void usage(void)
{
  version();
  discription();
  printf("Usage : fixbbl [options] filename[.bbl] [options]\n");
  printf("Options :\n");
  printf("\t-invisible\t convert to '0x7f0xa10x7f1230x7f'-format (default)\n");
  printf("\t-visible\t convert to '^^7f^^xx^^7fddd^^7f'-format\n");
  printf("\t-sort\n");
  printf("\t-sort=py\n");
  printf("\t-sort=pinyin\t sort all bibitems by PINYIN (ƴ) order\n");
  printf("\t-sort=bs\n");
  printf("\t-sort=bishun\t sort all bibitems by BISHUN (˳) order\n");
  printf("\t-sort=bhs\n");
  printf("\t-sort=bihuashu\t sort all bibitems by BIHUASHU (ʻ+˳) order\n");
  printf("\t-sort=uni\n");
  printf("\t-sort=unicode\t sort all bibitems by Unicode order\n");
  printf("\t-s\t\t run fixbbl silently (but errors remain)\n");
  printf("\t-npe\t\t disable display of all parsing errors\n");
  printf("\t-cjk\t\t CJK support (see below) are used\n");
  printf("\t-dispcjk\t display CJK support commands only\n");
  printf("\t-strings\t display commands on redefinable strings\n");
//  CJKcommands();
}

#define BIBPROCESSED	"%%& This .bbl file has been processed by fixbbl!\n"
#define BIBPROCESSED2	"%& This .bbl file has been processed by fixbbl!\n"

#define SORT_BY_NOTHING		0 // 
#define SORT_BY_PINYIN		1 // ƴ
#define SORT_BY_STROKES_O	2 // ˳A
#define SORT_BY_STROKES_N	3 // ʻB(ʻͬʱӦñ˳)
#define SORT_BY_UNICODE		4 // Unicode
BYTE	bSort = SORT_BY_NOTHING;

BYTE	bSilent = 0;
BYTE	bVisibleCJKchar = 0;
BYTE	bParsingErrors = 1;
BYTE	bCJKcircum = 0;
BYTE	bCJKprocessed = 0;
#define MAX_CHAR_LENGTH		256
//char	strAndChar[MAX_CHAR_LENGTH] = "";//'\&''and'滻
//ע⣬⣬еĻ¼֮andҲҪ滻
char	strAndChar[MAX_CHAR_LENGTH] = ", ";//'\&''and'滻
char	strEtAlChar[MAX_CHAR_LENGTH] = "";//'et al.'滻
char	strAndChar2[MAX_CHAR_LENGTH] = ", ";//'\&''and'滻'^^7f'
char	strEtAlChar2[MAX_CHAR_LENGTH] = "^^7f^^b5^^7f200^^7f";//'et al.'滻'^^7f'
//ԺֵĲתΪ'^^7f'ʽ
//ע⣬ļͻ¼ܳ֡, editorsҲҪ滻Ϊࡱ
char	str_EditorsChar[MAX_CHAR_LENGTH] = ", editors";//ߵEditorԭʼӢִ
char	str_EditorChar[MAX_CHAR_LENGTH] = ", editor";//ߵEditorԭʼӢִ
char	strEditorsChar[MAX_CHAR_LENGTH] = "~()";
//ڵڶ\newblockе²ַҪ滻
//һ\newblockеı⺬ĵ£滻
//˳δҵĽȱ:-(
char	str_PagesChar[MAX_CHAR_LENGTH] = "pages ";//'pages '滻
char	strPagesChar[MAX_CHAR_LENGTH] = "ҳ: ";//'pages '滻
char	str_ChapterChar[MAX_CHAR_LENGTH] = "chapter ";//'chapter '滻
char	strChapterChar[MAX_CHAR_LENGTH] = "";//'chapter '滻
char	str_EditionChar[MAX_CHAR_LENGTH] = " edition";//'Edition'滻
char	strEditionChar[MAX_CHAR_LENGTH] = "";//'Edition'滻
char	str_InChar[MAX_CHAR_LENGTH] = "In ";//'In '滻
char	strInChar[MAX_CHAR_LENGTH] = ": ";//'In '滻
char	str_Editors2Char[MAX_CHAR_LENGTH] = ", editors";//EditorԭʼӢִ
char	str_Editor2Char[MAX_CHAR_LENGTH] = ", editor";//EditorԭʼӢִ
char	strEditors2Char[MAX_CHAR_LENGTH] = "~()";//', editors'滻
char	endChar;//ǰеĬϽ']''\0'
FILE	*Fin, *Fout;

int		nLine;//ǰкţڶλ
//BYTE	bAuthorLine;//Ƿǰб
//BYTE	bBibLineFinished;//\bibitemǷ
//BYTE	nNewBlock;//ǰitem\newblock
//BYTE	bReplaceNewBlock;//滻ǰitem\newblockصݣ"In", "editors", "pages".

#define BIBPREFIX			"\\bibitem"
#define NEWBLOCKPREFIX		"\\newblock"
#define	LINE_EMPTY			0//ʼ״̬
#define	LINE_BIBITEM		1//'\bibitem'ĵһ
#define	LINE_BIBITEM2		111//'\bibitem'ĵһ('[]')
#define LINE_AUTHORS		2//'\bibitem'֮б
#define LINE_NEWBLOCK_TITLE	3//͵Ŀĵһnewblock
#define LINE_NEWBLOCK_BODY	4//newblockݵУһݣ
#define LINE_NEWBLOCK_MISC	5//newblockĸУΪУ
BYTE	bItemDataType;//ǰĿ
BYTE	bLastCJKChar;//һЧַǷΪַǿո
BYTE	bTeXCommand;//ǰַһTeXһ
BYTE	bChineseInProceedings;//ĿInProceedingsļ¼newblockҪת
BYTE	bInProcEditorsList;//ĿInProceedingsģλnewblockеıбҪ滻

void addSpaces(char *strSour)
{
	char strTemp[MAX_CHAR_LENGTH],*str,*str2;
	
	str = strSour;
	str2 = strTemp;
	while(*str!='\0') {
		if(*str == ' ') {
			while(*str==' ') str++;
			*str2++='~';
		}
		if(*str!=' ' && *str!='\0') *str2++=*str++;
	}
	*str2 = '\0';
	strcpy(strSour, strTemp);
}

//transform a CJK string to '^^7f'-format string
int CJK2Circum(char *strD, char *strS)
{
	BYTE h,l;
	char strTemp[MAX_CHAR_LENGTH];

	*strD = '\0';
	while(*strS!='\0') {
		h = (BYTE)*strS++;
		if(GBK_HIGH(h)) {
			l=*strS;
			if(GBK_LOW(l)) {
				strS++;
				if (bVisibleCJKchar) {
					if(l<100) sprintf(strTemp, "^^7f^^%x^^7f%2d^^7f", h, l);
					else sprintf(strTemp, "^^7f^^%x^^7f%3d^^7f", h, l);
				}
				else {
					if(l<100)
						sprintf(strTemp, "%c%c%c%2d%c", 0x7f, h, 0x7f, l, 0x7f);
					else
						sprintf(strTemp, "%c%c%c%3d%c", 0x7f, h, 0x7f, l, 0x7f);
				}
				strcat(strD, strTemp);
			}
			else {
				fprintf (stderr, "Wrong CJK characters in options:... ^^%x^^%x\n", h, l);
				return -1;
			}
		}
		else {
			sprintf(strTemp, "%c", h);
			strcat(strD, strTemp);
		}
	}
	return 1;
}

//write a Chinese character in Fout with '^^7f'-format string
int CJK2Circum2(BYTE h, BYTE l)
{
	if(GBK_HIGH(h)) {
		if(GBK_LOW(h)) {
			if (bVisibleCJKchar) {
				fprintf(Fout, "^^7f");
				fprintf(Fout, "^^%x^^7f", h);
				if(l<100) fprintf(Fout, "%2d^^7f", l);
				else fprintf(Fout, "%3d^^7f", l);
			}
			else {
				if(l<100)
					fprintf(Fout, "%c%c%c%2d%c", 0x7f, h, 0x7f, l, 0x7f);
				else
					fprintf(Fout, "%c%c%c%3d%c", 0x7f, h, 0x7f, l, 0x7f);
			}
		}
		else if (bParsingErrors) {
			fprintf (stderr, "An incomplete '^^xx' TeX character is found:\n\tLine %d: ... ^^%x^^%x\n", nLine, h, l);
			return -1;
		}
	}
	else fprintf(Fout, "%c%c", h, l);
	return 1;
}

//translate a '^^xx'-format TeX character to an ascii character
//if not a '^^xx'-format TeX character, return itself
int translateChar(char **str)
{
	BYTE a;

	if ( **str != '^') {
		a = **str;
		if ( **str != endChar && **str != '\0' && **str != '\n' && **str != '\r') (*str)++;
		return a;//if not '^^xx' directly return the current character
	}

	while(**str == '^') (*str)++;//skip all '^' characters
	if ( is16digit(**str) && is16digit(*(*str+1)) ) {
		a = xtoi(*str);
		(*str) = *str + 2;
		return a;
	}
	else {
		if (bParsingErrors)
			fprintf (stderr, "An incomplete '^^xx' TeX character is found:\n\tLine %d: ... \"%s\"\n", nLine, (*str)-2);
		return -1;//-1L = 0xffffffff
	}
}

//skip contunous spaces (maybe in '^^20' format)
int skipSpaces(char **str)
{
	int rtn;
	BYTE h;

	do{
		rtn = translateChar(str);
		if(rtn == -1) return -1;
		h = rtn;
	}while(h == ' ');
	
	return h;
}

//skip existent '^^7f111^^7f111^^7f' characters
int skipCircumCharacters(char **str)
{
	int rtn, i;
	BYTE h, l;
	char strTemp[20];//'^^7f111^^7f111^^7f'
	char *p = *str;

	if(isdigit(**str)) {
		for(i=0;i<3;i++) {
			if(!isdigit(**str)) break;
			strTemp[i] = **str;
			(*str)++;
		}
		strTemp[i]='\0';
		h = atoi(strTemp);
	}
	else {
		rtn = translateChar(str);
		if (rtn == -1) return 0;//break when encountering errors
		h = (BYTE) rtn;
	}
	rtn = translateChar(str);
	if (rtn != 0x7f) {
		if (bParsingErrors && rtn != -1)
			fprintf (stderr, "Wrong \"^^7f\" CJK character - \"^^7f\" is expected here:\n\tLine %d: ... ^^%x\n", nLine, rtn);
		return 0;//return 0 when encountering errors
	}
	if(isdigit(**str)) {
		for(i=0;i<3;i++) {
			if(!isdigit(**str)) break;
			strTemp[i] = **str;
			(*str)++;
		}
		strTemp[i]='\0';
		l = atoi(strTemp);
	}
	else {
		rtn = translateChar(str);
		if (rtn == -1) return 0;//break when encountering errors
		l = (BYTE) rtn;
	}
	rtn = translateChar(str);
	if (rtn != 0x7f) {
		if (bParsingErrors && rtn != -1)
			fprintf (stderr, "Wrong \"^^7f\" CJK character - \"^^7f\" is expected here:\n\tLine %d: ... ^^%x\n", nLine, rtn);
		return 0;//return 0 when encountering errors
	}
	fprintf(Fout, "^^7f");
	for(; p < *str; p++) fputc(*p, Fout);
	return 1;
}

//process 'et al.', '\&' and 'and' after at least one ' '
int processSpaceLeadingChars(char **str)
{
	int i, j, rtn;
	BYTE h, hh[5];

	bTeXCommand = 0;
	rtn = skipSpaces(str);
	if (rtn == -1) return -1;//continue when encountering errors
	if(!bLastCJKChar) {//for Non-Chinese names, directly return
		fputc(' ', Fout);
		return rtn;
	}
	h = (BYTE) rtn;
	if(h == '\\') {//is '\&'?
		rtn = translateChar(str);
		h = (BYTE) rtn;//when rtn == -1, h=0xFF != ' '
		if (rtn == -1) return -1;//continue when encountering errors
		if(h == '&') {
			if(bCJKprocessed || bItemDataType!=LINE_BIBITEM) fprintf(Fout, strAndChar);
			else fprintf(Fout, strAndChar2);
			rtn = skipSpaces(str);//skip all following spaces
			if (rtn == -1) return -1;//continue when encountering errors
			else return rtn;//do not continue for other cases
		}
		else {
			bTeXCommand = 1;//enter a TeX command
			fputc('\\', Fout);//do not continue for other characters!
		}
		return h;
	}
	else if(tolower(h) == 'a') {//is 'and'?
		bLastCJKChar = 0;
		for(i = 0; i < 2; i++) {
			rtn = translateChar(str);
			if (rtn == -1) break;
			hh[i] = (BYTE) rtn;//when rtn == -1, hh[i]=0xFF != 'nd'
			if (strchr("nd", tolower(hh[i])) == NULL) break;
		}
		if (rtn == -1) return -1;//continue when encountering errors"
		if(i >= 2) i--;//ȷȡ2ַiΪ±
		if(tolower(hh[0]) == 'n' && tolower(hh[1]) == 'd' ) {
			if(bCJKprocessed || bItemDataType!=LINE_BIBITEM) fprintf(Fout, strAndChar);
			else fprintf(Fout, strAndChar2);
			rtn = skipSpaces(str);//skip all following spaces
			bLastCJKChar = 1;
			if (rtn == -1) return -1;//continue when encountering errors
			else return rtn;//do not continue for other cases
		}
		else {
			fputc(h, Fout);
			if(strchr("nd", tolower(hh[i]))) {
				for(j = 0; j <= i; j++) fputc(hh[j], Fout);
				return -1;//continue if both are 'nd'
			}
			else {
				for(j = 0; j <= i-1; j++) fputc(hh[j], Fout);
				h = hh[i];
				//do not continue for other cases!
			}
		}
	}
	else if(tolower(h) == 'e') {//is 'et~al.' or 'et al.'? after CJK character
		bLastCJKChar = 0;
		for(i = 0; i < 5; i++) {
			rtn = translateChar(str);
			if (rtn == -1) break;
			hh[i] = (BYTE) rtn;//when rtn == -1, hh[i]=0xFF != 't~ al.'
			if (strchr("t~ al.", tolower(hh[i])) == NULL) break;
		}
		if (rtn == -1) return -1;//continue when encountering errors"
		if(i >= 5) i--;//ȷȡ5ַiΪ±
		if(tolower(hh[0])=='t' && hh[1]=='~' && tolower(hh[2])=='a' && tolower(hh[3])=='l' && hh[4]=='.') {
			if(bCJKprocessed || bItemDataType!=LINE_BIBITEM) fprintf(Fout, strEtAlChar);
			else fprintf(Fout, strEtAlChar2);
			bLastCJKChar = 1;
			rtn = skipSpaces(str);//skip all following spaces
			if (rtn == -1) return -1;//continue when encountering errors
			else return rtn;//do not continue for other cases
		}
		else if(tolower(hh[0])=='t' && hh[1]==' ' && tolower(hh[2])=='a' && tolower(hh[3])=='l' && hh[4]=='.') {
			if(bCJKprocessed || bItemDataType!=LINE_BIBITEM) fprintf(Fout, strEtAlChar);
			else fprintf(Fout, strEtAlChar2);
			bLastCJKChar = 1;
			rtn = skipSpaces(str);//skip all following spaces
			if (rtn == -1) return -1;//continue when encountering errors
			else return rtn;//do not continue for other cases
		}
		else {
			fputc(tolower(h), Fout);
			if(strchr("t~ al.", tolower(hh[i]))) {
				for(j = 0; j <= i; j++) fputc(hh[j], Fout);
				return -1;//continue if both are 'nd'
			}
			else {
				for(j = 0; j <= i-1; j++) fputc(hh[j], Fout);
				h = hh[i];//do not continue for other cases!
			}
		}
	}
	else fputc('~', Fout);
	return h;
}

//process ' and' after a ','
//avoid cancel ' ' after ','
int processAndAfterComma(char **str)
{
	int i, j, rtn;
	BYTE h, hh[5];

	bTeXCommand = 0;
	if(!bLastCJKChar) {
		fputc(',', Fout);
		return -1;//continue if the left character is ','
	}
	bLastCJKChar = 0;
	for(i = 0; i < 4; i++) {
		rtn = translateChar(str);
		if (rtn == -1) break;
		hh[i] = (BYTE) rtn;//when rtn == -1, hh[i]=0xFF != ' and'
		if (strchr(" and", tolower(hh[i])) == NULL) break;
	}
	if (rtn == -1) return -1;//continue when encountering errors"
	if(i >= 4) i--;//ȷȡ4ַiΪ±
	if(hh[0]==' ' && tolower(hh[1])=='a' && tolower(hh[2])=='n' && tolower(hh[3])=='d') {
		if(bCJKprocessed || bItemDataType!=LINE_BIBITEM) fprintf(Fout, strAndChar);
		else fprintf(Fout, strAndChar2);
		bLastCJKChar = 1;
		rtn = skipSpaces(str);//skip all following spaces
		if (rtn == -1) return -1;//continue when encountering errors
		else return rtn;//do not continue for other cases
	}
	fputc(',', Fout);//the first character must be ','
	if(strchr(" and", tolower(hh[i]))) {
		for(j = 0; j <= i; j++) fputc(hh[j], Fout);
		return -1;//continue if both are ' and'
	}
	else {//do not continue for other cases!
		//please note that i = the actual number of read characters
		for(j = 0; j <= i-1; j++) {
			if(hh[j] == '\\') bTeXCommand = 1;
			else if(!bTeXCommand && isalpha(hh[j])) bLastCJKChar = 0;//only when !bTeXCommand non-Chinese is end
			else if(bTeXCommand && !isalpha(hh[j])) bTeXCommand = 0;//only when a non-letter occurrs, TeXCommand is end
			fputc(hh[j], Fout);
		}
		h = hh[i];
	}
	return h;
}

//simply copy all common and CJK characters when CJK commands are used
void copyCJKstring(BYTE *h, char **str)
{
	int rtn;
	BYTE l;

	if(GBK_HIGH(*h)) {
		rtn = translateChar(str);
		if (rtn == -1) return;//return when encountering errors
		else l = (BYTE) rtn;
		if(GBK_LOW(l)) {
			bLastCJKChar = 1;
			fprintf(Fout, "%c%c", *h, l);
		}
		else {
			if (bParsingErrors)
				fprintf (stderr, "A wrong CJK character is found:\n\tLine %d: ... ^^%x^^%x\n", nLine, *h, l);
			*h = l;
		}
	}
	else if (*h!='\0' && *h!=endChar && *h!='\n' && *h!='\r') {
		if(*h == '\\') bTeXCommand = 1;
		else if(!bTeXCommand && isalpha(*h)) bLastCJKChar = 0;//only when !bTeXCommand non-Chinese is end
		else if(bTeXCommand && !isalpha(*h)) bTeXCommand = 0;//only when a non-letter occurrs, TeXCommand is end
		fputc(*h, Fout);
	}
}

char			*b_in = NULL;
char			*p, *b2_in, *b2_in_start;
char			inname[PATH_MAX]="";//.bib file name
char			outname[PATH_MAX]="";//.bib.tmp file name
char			bakname[PATH_MAX]="";//.bib.bak file name and also for temp use
char			texname[PATH_MAX]="";//.tex file name
unsigned int	nBibFileSize;//size of .bib file
int				rtn, i;
BYTE			h, l;

int openbblfile()
{
	Fin = fopen(inname, "r");
	if(!Fin) {
		fprintf(stderr, "Failed to open %s for reading!\n", inname);
		return 2;
	}
	nBibFileSize = _filelength (_fileno(Fin));
	if (nBibFileSize == -1L) {
		fprintf(stderr, "Cannot get the file size of %s!\n", inname);
		fclose (Fin);
		return 3;
	}
	b_in = (unsigned char *)malloc(nBibFileSize);
	if(!b_in) {
		fprintf(stderr, "Memory allocation error!\n");
		fclose (Fin); 
		return 4;
	}
	return 0;
}

int IsFileExist(char *strFile)
{
#ifdef WIN32
	struct _finddata_t fileinfo;
	int		hFileFind;
	hFileFind = _findfirst(strFile, &fileinfo);
	if(hFileFind == -1 && errno == ENOENT) return 0;
	else {
		_findclose(hFileFind);
		return 1;
	}
#else//UNIX
    glob_t g;
	if(glob(strFile, GLOB_MARK | GLOB_PERIOD, NULL, &g)) return 0;
	globfree(&g);
	return 1;
#endif
}

int CheckBBLFile(BYTE falg)
{
	BYTE bBBLFileExist;
	if(falg) bBBLFileExist = 0;//when flag = 1, bibtex foo must be called
	else
	bBBLFileExist = IsFileExist(inname);
	if(!bBBLFileExist) {
		if(!bSilent) fprintf(stdout, ">>>%s...\n", bakname);
		system(bakname);//call bibtex foo[.aux] to generate foo.bbl
		if(!bSilent) fprintf(stdout, "<<<%s is finished!\n", bakname);
		bBBLFileExist = IsFileExist(inname);
		if(!bBBLFileExist) return 1;
	}
	rtn = openbblfile();//try to open .bib file again
	if(rtn != 0) return rtn;
	return 0;
}

int openBBLFile()
{
	BYTE bChecked = 0;
	int  rtn;

	sprintf(bakname, "bibtex %s", texname);
	rtn = CheckBBLFile(0);
	if(rtn!=0) return rtn;
	if(fgets(b_in, nBibFileSize, Fin) == NULL) {
		fclose(Fin);
		rtn = CheckBBLFile(1);
		if(rtn!=0) return rtn;
		if(fgets(b_in, nBibFileSize, Fin) == NULL) {
			fclose(Fin);
			return 2;
		}
	}
	if(!strcmp(b_in, BIBPROCESSED2)) {
		fclose(Fin);
		rtn = CheckBBLFile(1);
		if(rtn!=0) return rtn;
	}
	else rewind(Fin);
	return 0;
}

char fixbblPath[PATH_MAX];

//Find and Open strFileName in current PATH
//this function can be simplified with _searchenv()
FILE *OpenFileInPATH(char *strFileName, char *strMode)
{
	char *strPath;
	FILE *Ftemp;

	//Find and Open strFileName in current path
	Ftemp = fopen(strFileName, strMode);
#ifdef _DEBUG
	if(Ftemp != NULL) printf("%s is found in the current path!\n", strFileName);
	else printf("%s does not exist in the current path!\n", strFileName);
#endif
	if(Ftemp != NULL) return Ftemp;
	
	//Find and Open strFileName in PATH of fixbbl.exe
	if(fixbblPath[0] != '\0') {
		sprintf(bakname, "%s%s", fixbblPath, strFileName);
		Ftemp = fopen(bakname, strMode);
		if(Ftemp != NULL) return Ftemp;
	}

	//Find and Open strFileName in paths@PATH
	strPath = getenv("Path");
	while(*strPath != '\0') {
		p = strchr(strPath, ';');
		if(p != NULL || *strPath != '\0')
		strncpy(bakname, strPath, p - strPath);
		if(*(p-1) == '\\') bakname[p - strPath] = '\0';
//		else if(*(p-1) == '.') bakname[p - strPath - 1] = '\0';
		else {
			bakname[p - strPath] = '\\';
			bakname[p - strPath + 1] = '\0';
		}
		strcat(bakname, strFileName);
		Ftemp = fopen(bakname, strMode);
		if(Ftemp != NULL) return Ftemp;
		if(p == NULL) break;
		strPath = ++p;
	}
	return NULL;
}

void Editors2EditorsShort(char *strEditorsShort, char *strEditors)
{
	char *p = strEditors;
	while(*p==' ' || *p==',') p++;
	strcpy(strEditorsShort, p);
}

void loadCFG()
{
	FILE *Fcfg;
	int i,j;
	if((Fcfg = OpenFileInPATH("fixbbl.cfg", "rt")) == NULL) return;
	while(!feof(Fcfg)) {
		if(fgets(bakname, PATH_MAX, Fcfg) == NULL) break;
		if(bakname[0] == '%') continue;//comments
		i=strlen(bakname);
		for(j=0;j<i;j++) {
			if(bakname[j] == '\n' || bakname[j] == '\r') {
				bakname[j]='\0';break;
			}
		}
		if (!strnicmp(bakname, "and=",4)) strcpy(strAndChar,bakname+4);
		else if (!strnicmp(bakname, "etal.=",6)) strcpy(strEtAlChar,bakname+6);
		else if (!strnicmp(bakname, "_In=",4)) strcpy(str_InChar,bakname+4);
		else if (!strnicmp(bakname, "In=",3)) strcpy(strInChar,bakname+3);
		else if (!strnicmp(bakname, "_pages=",7)) strcpy(str_PagesChar,bakname+7);
		else if (!strnicmp(bakname, "pages=",6)) strcpy(strPagesChar,bakname+6);
		else if (!strnicmp(bakname, "_chapter=",9)) strcpy(str_ChapterChar,bakname+9);
		else if (!strnicmp(bakname, "chapter=",8)) strcpy(strChapterChar,bakname+8);
		else if (!strnicmp(bakname, "_edition=",9)) strcpy(str_EditionChar,bakname+9);
		else if (!strnicmp(bakname, "edition=",8)) strcpy(strEditionChar,bakname+8);
		else if (!strnicmp(bakname, "_editors=",9)) strcpy(str_EditorsChar,bakname+9);
		else if (!strnicmp(bakname, "_editor=",8)) strcpy(str_EditorChar,bakname+8);
		else if (!strnicmp(bakname, "editors=",8)) strcpy(strEditorsChar,bakname+8);
		else if (!strnicmp(bakname, "_editors2=",10)) strcpy(str_Editors2Char,bakname+10);
		else if (!strnicmp(bakname, "_editor2=",9)) strcpy(str_Editor2Char,bakname+9);
		else if (!strnicmp(bakname, "editors2=",9)) strcpy(strEditors2Char,bakname+9);
	}
	fclose(Fcfg);
}

#include "GBK_sort.c"

int main(int argc, char* argv[])
{
#ifdef _DEBUG
	// Unicode -- : 0x5f20, : 0x4e09, : 0x56db
/*	char strTemp1[512], strTemp2[512];
	int result;
	// bSort = SORT_BY_PINYIN;
	// getGBK2pyInfo();
	ProcessAuthors(strTemp1, "Shujun Li, Yuanlong Cai, and Ĳ.");
	ProcessAuthors(strTemp2, "Shujun Li, Ĳ and Yuanlong Cai.");
	result = strcmp_pinyin_Authors(strTemp1, strTemp2, 1, 2);
	ProcessAuthors(strTemp2, "Si and {John (Wiley) \\& Sons, Inc.}, et al.");
//	Authors2Authors(strTemp2, "Peter and John, Ltd., \\pinyin{shu4} (Hook) Li, editors");
	result = strcmp_strokes_Authors(strTemp1, strTemp2, 1, 2);
	Authors2Authors(strTemp1, ", , S. Zhang, Li, S.,  and Wang, W., editors.xxxxx");
	Authors2Py(strPinyin1, strTemp1);
	Authors2Authors(strTemp2, " (S. Zhang) \\& , S.~Zhang (), Li, S., , \\& Wang, W. [2000] xxxx");
	Authors2Py(strPinyin2, strTemp2);
	result = strcmp_pinyin_Authors(strPinyin1, strPinyin2, 1, 2);
	result = strcmp_Unicode("", "", GBK_LESS_NON_GBK);
	result = strcmp_Unicode("", "San Zhang", GBK_LESS_NON_GBK);
	result = strcmp_Unicode("", "San Zhang", NON_GBK_LESS_GBK);
	result = strcmp_Unicode(" San", " Si", GBK_LESS_NON_GBK);
	result = strcmp_Unicode("Sin", "Si", NON_GBK_LESS_GBK);
	result = strcmp_Unicode("Sin", "Si", GBK_LESS_NON_GBK);
	result = strcmp_Unicode_GBK("", "");
	result = strcmp_Unicode_GBK("", "");
	Authors2Authors(strTemp1, ", , S. Zhang, Li, S.,  and Wang, W., editors.xxxxx");
	Authors2Authors(strTemp2, " (S. Zhang) \\& , S.~Zhang (), Li, S., , \\& Wang, W. [2000] xxxx");
	result = strcmp_Unicode_Authors(strTemp1, strTemp2, 1, 2);
	Authors2Authors(strTemp, ", , S. Zhang, Li, S., , Wang, W., et al.xxxxxx");
	Authors2Authors(strTemp, " \\& , San John Zhang, Li, S., , Wang, W. [2000] xxxx");
	Authors2Authors(strTemp, " \\& , S.~Zhang, Li, S., , Wu Wang (2000) xxxx");
	Authors2Authors(strTemp, " \\& , S.~Zhang, Li, S., , \\& Wu Wang [2000] xxxx");
	Authors2Authors(strTemp, " (S. Zhang) \\& , S.~Zhang (), Li, S., , \\& Wu Wang [2000] xxxx"); */
	// exit(1);
#endif

	p = strrchr(argv[0], '\\');
	if(p != NULL) {
		strncpy(fixbblPath, argv[0], p - argv[0] + 1);
		fixbblPath[p - argv[0] + 1] = '\0';
	}
	loadCFG();

	for(i=1; i < argc; i++) {
		#if defined(WIN32) || defined(GO32)
			strlwr(argv[i]);
		#endif
		if (!strcmp(argv[i], "-sort")) bSort = SORT_BY_PINYIN; //ĬʽΪƴ
		else if (!strnicmp(argv[i], "-sort=",6)) {
			if(!strnicmp(argv[i]+6, "pinyin", 6)) bSort = SORT_BY_PINYIN;
			else if(!strnicmp(argv[i]+6, "py", 2)) bSort = SORT_BY_PINYIN;
			else if(!strnicmp(argv[i]+6, "bishun", 5)) bSort = SORT_BY_STROKES_O;
			else if(!strnicmp(argv[i]+6, "bs", 2)) bSort = SORT_BY_STROKES_O;
			else if(!strnicmp(argv[i]+6, "bihuashu", 5)) bSort = SORT_BY_STROKES_N;
			else if(!strnicmp(argv[i]+6, "bhs", 5)) bSort = SORT_BY_STROKES_N;
			else if(!strnicmp(argv[i]+6, "unicode", 2)) bSort = SORT_BY_UNICODE;
			else if(!strnicmp(argv[i]+6, "uni", 2)) bSort = SORT_BY_UNICODE;
			else bSort = SORT_BY_NOTHING;
		}
		else if (!strcmp(argv[i], "-s")) bSilent = 1;
		else if (!strcmp(argv[i], "-npe")) bParsingErrors = 0;
		else if (!strcmp(argv[i], "-invisible")) bVisibleCJKchar = 0;
		else if (!strcmp(argv[i], "-visible")) bVisibleCJKchar = 1;
		else if (!strcmp(argv[i], "-cjk")) bCJKprocessed = 1;
		else if (!strcmp(argv[i], "-dispcjk")) {
			version();
			CJKcommands();
			exit(1);
		}
		else if (!strcmp(argv[i], "-strings")) {
			RedefinableStrings();
			exit(1);
		}
		else if (!strnicmp(argv[i], "-and=",5)) strcpy(strAndChar,argv[i]+5);
		else if (!strnicmp(argv[i], "-etal.=",7)) strcpy(strEtAlChar,argv[i]+7);
		else if (!strnicmp(argv[i], "-_In=",5)) strcpy(str_InChar,argv[i]+5);
		else if (!strnicmp(argv[i], "-In=",4)) strcpy(strInChar,argv[i]+4);
		else if (!strnicmp(argv[i], "-_editors=",10)) strcpy(str_EditorsChar,argv[i]+10);
		else if (!strnicmp(argv[i], "-_editor=",9)) strcpy(str_EditorChar,argv[i]+9);
		else if (!strnicmp(argv[i], "-editors=",9)) strcpy(strEditorsChar,argv[i]+9);
		else if (!strnicmp(argv[i], "-_editors2=",11)) strcpy(str_Editors2Char,argv[i]+11);
		else if (!strnicmp(argv[i], "-_editor2=",10)) strcpy(str_Editor2Char,argv[i]+10);
		else if (!strnicmp(argv[i], "-editors2=",10)) strcpy(strEditors2Char,argv[i]+10);
		else if (!strnicmp(argv[i], "-_pages=",8)) strcpy(str_PagesChar,argv[i]+8);
		else if (!strnicmp(argv[i], "-pages=",7)) strcpy(strPagesChar,argv[i]+7);
		else if (!strnicmp(argv[i], "-_chapter=",10))strcpy(str_ChapterChar,argv[i]+10);
		else if (!strnicmp(argv[i], "-chapter=",9)) strcpy(strChapterChar,argv[i]+9);
		else if (!strnicmp(argv[i], "-_edition=",10)) strcpy(str_EditionChar,argv[i]+10);
		else if (!strnicmp(argv[i], "-edition=",9)) strcpy(strEditionChar,argv[i]+9);
		else strcpy(inname, argv[i]);
	}

	addSpaces(strAndChar);
	CJK2Circum(strAndChar2, strAndChar);
	addSpaces(strEtAlChar);
	CJK2Circum(strEtAlChar2, strEtAlChar);

	if(inname[0] == '\0') {
		usage();
		return 1;
	}

	if (!bSilent) version();//display version and developer information

	p = strrchr(inname, '.');
	if(p != NULL) {
		strncpy(bakname, inname, p-inname);
		bakname[p-inname] = '\0';
		strcpy(texname, bakname);
	}
	else strcpy(texname, inname);
	#ifdef WIN32
		if((p == NULL) || stricmp(p, ".bbl")) strcat(inname, ".bbl");
	#else
		if((p == NULL) || strcmp(p, ".bbl")) strcat(inname, ".bbl");
	#endif

	strcpy(outname, inname);
	strcat(outname, ".tmp");

	if(openBBLFile()!=0) {
		free(b_in); exit(1);
	}

	Fout = fopen(outname, "wt");
	if(!Fout) {
		fprintf(stderr, "Cannot open %s for writing!\n", outname);
		free(b_in); fclose(Fin); exit(1);
	}

	if(bSort != 0) {
		if(SortBibitems() == -1) exit(1);
	}

	fprintf(Fout, BIBPROCESSED);//write the flag into .bbl file.

	nLine = 0;
//	bAuthorLine = 0;
//	bBibLineFinished = 1;
	bItemDataType = LINE_EMPTY;
	bLastCJKChar = 0;
	bTeXCommand = 0;
	bChineseInProceedings = 0;
	bInProcEditorsList = 0;
//	bReplaceNewBlock = 0;
//	nNewBlock = 0;
	while(!feof(Fin)) {
		if(fgets(b_in, nBibFileSize, Fin) == NULL) break;
		nLine++;
		b2_in = b_in;
		//if(bAuthorLine || !bBibLineFinished) {
		//skip all leading spaces before each line
		while (*b2_in == ' ') b2_in++;
		//}
		//There cannot be more than one \bibtem in a line!!
		//if(!bAuthorLine && bBibLineFinished) {
			//find '\bibitem' only when the last '\bibitem' finished
			//and the current line is not authors list
		//skip any empty line
		if((*b2_in=='\n' && *(b2_in+1)=='\r') || (*b2_in=='\r' && *(b2_in+1)=='\n')) {
			bItemDataType = LINE_EMPTY;
			fprintf(Fout, "%c%c", *b2_in++, *b2_in++);
			continue;
		}
		//find '\bibitem' line
		p = strstr(b2_in, BIBPREFIX);
/*			if( p == NULL) {
				fprintf(Fout, "%s", b2_in);//copy the current line
				bAuthorLine = 0;//the following line will not be authors list
				continue;
			}*/
		if( p != NULL) {
			bItemDataType = LINE_BIBITEM;
			bChineseInProceedings = 0;
//			bAuthorLine = 0;
//			bBibLineFinished = 0;//the \bibitem line has not been finished
//			bReplaceNewBlock = 0;
//			nNewBlock = 0;
			for(; b2_in < p; b2_in++) fputc(*b2_in, Fout);//copy the left content
			fprintf(Fout, BIBPREFIX);
			b2_in = p + strlen(BIBPREFIX);//skip '\bibitem'
			//copy the contents between '\bibitem' and the first '['
			while (*b2_in !='\0' && *b2_in != '[') {
				if(*b2_in == '{') bItemDataType = LINE_BIBITEM2;
				fputc(*b2_in, Fout);
				b2_in++;
			}
		}
		else {//if '\bibitem' is not found, try to find '\newblock'
			p = strstr(b2_in, NEWBLOCKPREFIX);
			if( p != NULL) {
				for(; b2_in < p; b2_in++) fputc(*b2_in, Fout);//copy the left content
				fprintf(Fout, NEWBLOCKPREFIX);
				b2_in = p + strlen(NEWBLOCKPREFIX);//skip '\bibitem'
				//copy the contents between '\bibitem' and the first non-space character
				while (*b2_in == ' ') {
					fputc(*b2_in, Fout);
					b2_in++;
				}
				switch(bItemDataType) {
					case LINE_BIBITEM:
					case LINE_AUTHORS: bItemDataType = LINE_NEWBLOCK_TITLE; break;
					case LINE_NEWBLOCK_TITLE: bItemDataType = LINE_NEWBLOCK_BODY; break;
					case LINE_NEWBLOCK_BODY: bItemDataType = LINE_NEWBLOCK_MISC; break;
					default: bItemDataType = LINE_NEWBLOCK_MISC;
				}
				bInProcEditorsList = 0;
			}
		}
//		}
		if(bItemDataType == LINE_EMPTY || bItemDataType == LINE_NEWBLOCK_MISC 
			|| (bItemDataType == LINE_NEWBLOCK_BODY && !bChineseInProceedings)) {
			//copy the current line and continue;
			fprintf(Fout, "%s", b2_in);
			continue;
		}
		if(bItemDataType == LINE_NEWBLOCK_TITLE) {
			//copy the current line and continue;
//			fprintf(Fout, "%s", b2_in);
			b2_in_start = b2_in;
			while(*b2_in!='\0' && *b2_in!='\n' && *b2_in!='\r') {
				rtn = translateChar(&b2_in);
				if (rtn == -1) break;//break when encountering errors
				else h = (BYTE) rtn;
				if(GBK_HIGH(h)) {
					bChineseInProceedings = 1;
					break;
				}
			}
			b2_in = b2_in_start;
//			continue;
		}
		if(bItemDataType != LINE_BIBITEM) endChar = '\0';
		else endChar = ']';
		h = *b2_in;
		b2_in_start = b2_in;//save the current b2_in
		while(h!='\0' && h!=endChar && h!='\n' && h!='\r') {//do untill ']', '\0', '\n' or '\r'
			//if 'In ' occurrs in the beginning of LINE_NEWBLOCK_BODY
			//then editors lists may occur
			if(bItemDataType == LINE_NEWBLOCK_BODY && b2_in==b2_in_start) {
				if(!strncmp(b2_in, str_InChar, strlen(str_InChar)))
				{
					fprintf(Fout, "%s", strInChar);
					b2_in += strlen(str_InChar);
					bInProcEditorsList = 1;
					continue;
				}
				if(str_EditionChar[0]==' ') {
					if(!strncmp(b2_in, str_EditionChar+1, strlen(str_EditionChar)-1)) {
						fprintf(Fout, "%s", strEditionChar);
						b2_in += strlen(str_EditionChar+1);
						continue;
					}
				}
			}
			rtn = translateChar(&b2_in);
			if (rtn == -1) break;//break when encountering errors
			else h = (BYTE) rtn;
			if(bItemDataType == LINE_NEWBLOCK_BODY || bItemDataType == LINE_NEWBLOCK_TITLE) {
				if(h==str_PagesChar[0] && !strncmp(b2_in, str_PagesChar+1, strlen(str_PagesChar)-1)) {
					fprintf(Fout, "%s", strPagesChar);
					b2_in += strlen(str_PagesChar+1);
					continue;
				}
				if(h==str_ChapterChar[0] && !strncmp(b2_in, str_ChapterChar+1, strlen(str_ChapterChar)-1)) {
					fprintf(Fout, "%s", strChapterChar);
					b2_in += strlen(str_ChapterChar+1);
					continue;
				}
			}
			if(strstr(b2_in, "ditor")) bInProcEditorsList = 1; // if "Editor(s)" or "editor(s)" occurs
			if(bItemDataType == LINE_NEWBLOCK_BODY) {
				//if '{\em ...}' occurrs, editors list should be end
				if(h=='{') bInProcEditorsList = 0;
				//When ', editor(s), ' occurrs in LINE_NEWBLOCK_BODY ...
				if(bInProcEditorsList) {
					if(h==str_Editors2Char[0] || h==str_Editor2Char[0]) {
						if(h==str_Editors2Char[0] && !strncmp(b2_in, str_Editors2Char+1, strlen(str_Editors2Char)-1)) {
							fprintf(Fout, "%s", strEditors2Char);
							b2_in += strlen(str_Editors2Char+1);
							bInProcEditorsList = 0;//editors list is end in LINE_NEWBLOCK_BODY
							continue;
						}
						else if(h==str_Editor2Char[0] && !strncmp(b2_in, str_Editor2Char+1, strlen(str_Editor2Char)-1)) {
							fprintf(Fout, "%s", strEditors2Char);
							b2_in += strlen(str_Editor2Char+1);
							bInProcEditorsList = 0;//editors list is end in LINE_NEWBLOCK_BODY
							continue;
						}
					}
				}
				if(h==str_EditionChar[0] && !strncmp(b2_in, str_EditionChar+1, strlen(str_EditionChar)-1)) {
					fprintf(Fout, "%s", strEditionChar);
					b2_in += strlen(str_EditionChar+1);
					continue;
				}
			}
			//When ', editor(s)' occurrs in LINE_AUTHORS ...
			// ', editor(s)' may occurr otherwhere for proceeding papers
			if(bItemDataType == LINE_AUTHORS) {
				if(h==str_EditorsChar[0] || h==str_EditorChar[0]) {
					if(h==str_EditorsChar[0] && !strncmp(b2_in, str_EditorsChar+1, strlen(str_EditorsChar)-1)) {
						fprintf(Fout, "%s", strEditorsChar);
						b2_in += strlen(str_EditorsChar+1);
//						bItemDataType = LINE_NEWBLOCK_TITLE;
//						bChineseInProceedings = 0;//The current item is "@Proceedings"
						continue;
					}
					else if(h==str_EditorChar[0] && !strncmp(b2_in, str_EditorChar+1, strlen(str_EditorChar)-1)) {
						fprintf(Fout, "%s", strEditorsChar);
						b2_in += strlen(str_EditorChar+1);
						continue;
					}
				}
			}
			//change ' ' to '~' and find possible 'and', '\&' and 'et~al.'
			//for the bibitem line and authors-list line
			if((bItemDataType == LINE_BIBITEM || bItemDataType == LINE_AUTHORS 
				|| (bItemDataType == LINE_NEWBLOCK_BODY && bInProcEditorsList))
				&& bLastCJKChar) {
				//replace ', editors' and set related flag
				//to simplify codes, I assume ', editors' can also occur in '\bibitem' line
				if(h == ' ') {
					rtn = processSpaceLeadingChars(&b2_in);
					if(rtn == -1) continue;//-1 means continue;
					h = rtn;
				}
				else if(h == ',') {//is ', and'?
					rtn = processAndAfterComma(&b2_in);
					if(rtn == -1) continue;//-1 means continue;
					h = rtn;
				}
			}
			//if CJK commands are used, simply copy all CJK characters
			if((bItemDataType == LINE_BIBITEM && bCJKprocessed) || bItemDataType!=LINE_BIBITEM) {
				copyCJKstring(&h, &b2_in);
				continue;//continue after copy a CJK character
			}
			//skip all '^^7f' characters
			if(h == 0x7f) {//encountering '^^7f...'
				if(skipCircumCharacters(&b2_in) == -1) {
					bLastCJKChar = 0; bTeXCommand = 0;
				}
				else bLastCJKChar = 1;
				continue;
			}
			if(GBK_HIGH(h)) {
				rtn = translateChar(&b2_in);
				if (rtn == -1) continue;//break when encountering errors
				else l = (BYTE) rtn;
				if(GBK_LOW(l)) {
					bLastCJKChar = 1;
					CJK2Circum2(h, l);
				}
				else {
					if (bParsingErrors)
						fprintf (stderr, "An incomplete '^^xx' TeX character is found:\n\tLine %d: ... ^^%x^^%x\n", nLine, h, l);
					h = l;
				}
			}
			else if (h!='\0' && h!=endChar && h!='\n' && h!='\r') {
				if(h == '\\') bTeXCommand = 1;
				else if(!bTeXCommand && isalpha(h)) bLastCJKChar = 0;//only when !bTeXCommand non-Chinese is end
				else if(bTeXCommand && !isalpha(h)) bTeXCommand = 0;//only when a non-letter occurrs, TeXCommand is end
				fputc(h, Fout);//normal ASCII characters
			}
		}
		//copy the left contents in the current line and set related flags
//		if(bAuthorLine) bAuthorLine = 0;//reset tag for authors list
		//the following line will be authors list
		if((bItemDataType == LINE_BIBITEM && h == ']') || bItemDataType == LINE_BIBITEM2)
			bItemDataType = LINE_AUTHORS;
		while(*b2_in!='\0' && *b2_in!='\n' && *b2_in!='\r')
			fputc(*b2_in++, Fout);
		if(*b2_in=='\n' || *b2_in=='\r') {
			if((bItemDataType == LINE_BIBITEM || bItemDataType == LINE_NEWBLOCK_BODY) && bLastCJKChar)
				fputc('%',Fout);
		}
		while(*b2_in!='\0') fputc(*b2_in++, Fout);
	}

	freeall();
	fclose(Fin);
	fclose(Fout);

	sprintf(bakname, "%s.bak",inname);
	remove(bakname);
	rename(inname,bakname);
	rename(outname,inname);

	if (!bSilent) fprintf(stdout, "fixbbl %s is finished!\n",inname);

//	getchar();
	return 0;

}

